open Mltype
open Printf

(****************************************************************************************************)

let parse fichier =
  let aparser = open_in fichier and grou = ref [] in
  try
    let lexbuf = Lexing.from_channel aparser in
    while true do
      let result = Parser.letrec Lexer.token lexbuf in
      grou := result :: !grou
    done; assert false
  with
    | Lexer.Eof -> close_in aparser; List.rev !grou

(****************************************************************************************************)

module Ident = Set.Make(String)

let var_libres =
  let rec vars accu = function
    | VAR x -> Ident.add x accu
    | APP (a, b) -> vars (vars accu a) b
    | FUN (f, x) -> Ident.union accu (Ident.remove f (vars Ident.empty x))
    | LET (a, x, e) -> Ident.union (vars accu x) (Ident.remove a (vars Ident.empty e))
    | REF a -> vars accu a
    | BANG a -> vars accu a
    | SET (a, b) -> vars (vars accu a) b
    | INT i -> accu
    | PLUS (a, b) -> vars (vars accu a) b
    | MINUS (a, b) -> vars (vars accu a) b
    | TIMES (a, b) -> vars (vars accu a) b
    | DIV (a, b) -> vars (vars accu a) b
    | REM (a, b) -> vars (vars accu a) b
    | OPP a -> vars accu a
    | LESS (a, b) -> vars (vars accu a) b
    | EQUAL (a, b) -> vars (vars accu a) b
    | TUPLE liste -> List.fold_left vars accu liste
    | PROJ (a, b) -> vars (vars accu a) b
    | SEQ (a, b) -> vars (vars accu a) b
    | IF (a, b, c) -> vars (vars (vars accu a) b) c
    | CALLPRIM (f, liste) -> List.fold_left vars accu liste
    | UNIT () -> accu
    | Op s -> accu
  in vars Ident.empty

(****************************************************************************************************)

let string_of_list f =
  let rec aux = function
    | [] -> " ]"
    | a :: q -> "; " ^ (f a) ^ (aux q)
  in function
    | [] -> "[]"
    | a :: q -> "[ " ^ (f a) ^ (aux q)

let rec string_of_expr = function
  | VAR a -> sprintf "(VAR \"%s\")" (String.escaped a)
  | APP (a, b) -> sprintf "(APP (%s, %s))" (string_of_expr a) (string_of_expr b)
  | FUN (a, b) -> sprintf "(FUN (\"%s\", %s))" (String.escaped a) (string_of_expr b)
  | LET (a, b, c) ->
      sprintf "(LET (\"%s\", %s, %s))" (String.escaped a) (string_of_expr b) (string_of_expr c)
  | REF a -> sprintf "(REF %s)" (string_of_expr a)
  | BANG a -> sprintf "(BANG %s)" (string_of_expr a)
  | SET (a, b) -> sprintf "(SET (%s, %s))" (string_of_expr a) (string_of_expr b)
  | INT a -> sprintf "(INT %d)" a
  | PLUS (a, b) -> sprintf "(PLUS (%s, %s))" (string_of_expr a) (string_of_expr b)
  | MINUS (a, b) -> sprintf "(MINUS (%s, %s))" (string_of_expr a) (string_of_expr b)
  | TIMES (a, b) -> sprintf "(TIMES (%s, %s))" (string_of_expr a) (string_of_expr b)
  | DIV (a, b) -> sprintf "(DIV (%s, %s))" (string_of_expr a) (string_of_expr b)
  | REM (a, b) -> sprintf "(REM (%s, %s))" (string_of_expr a) (string_of_expr b)
  | OPP a -> sprintf "(OPP %s)" (string_of_expr a)
  | LESS (a, b) -> sprintf "(LESS (%s, %s))" (string_of_expr a) (string_of_expr b)
  | EQUAL (a, b) -> sprintf "(EQUAL (%s, %s))" (string_of_expr a) (string_of_expr b)
  | TUPLE liste -> sprintf "(TUPLE %s)" (string_of_list string_of_expr liste)
  | PROJ (a, b) -> sprintf "(PROJ (%s, %s))" (string_of_expr a) (string_of_expr b)
  | SEQ (a, b) -> sprintf "(SEQ (%s, %s))" (string_of_expr a) (string_of_expr b)
  | IF (a, b, c) ->
      sprintf "(IF (%s, %s, %s))" (string_of_expr a) (string_of_expr b) (string_of_expr c)
  | CALLPRIM (f, liste) ->
      sprintf "(CALLPRIM (\"%s\", %s))" (String.escaped f) (string_of_list string_of_expr liste)
  | UNIT () -> "(UNIT ())"
  | Op a -> sprintf "(Op \"%s\")" (String.escaped a)

(****************************************************************************************************)
