open Mltype
open Printf

(******************************************************************************************)

type constructeur =
  | Base of string
  | Fun of constructeur * constructeur
  | Tuple of constructeur
  | Ref of constructeur
  | Var of int

module TypingEnv = Map.Make(String)
module Integer =
  struct
    type t = int
    let compare = Pervasives.compare
  end
module Substitution = Map.Make(Integer)
module VarSet = Set.Make(Integer)

exception Occur_check of int * string
exception Clash of string * string
exception Not_supported
exception Unbound of string

(******************************************************************************************)

let compteur =
  let x = ref 10 in
  fun () -> incr x; !x

let rec to_string = function
  | Base s -> s
  | Fun (a, b) -> sprintf "(%s -> %s)" (to_string a) (to_string b)
  | Tuple a -> sprintf "%s tuple" (to_string a)
  | Ref   a -> sprintf "%s ref"   (to_string a)
  | Var   x -> sprintf "%d" x

(******************************************************************************************)

let rec mem v = function
  | Var i when v = i -> true
  | Fun (a, b) -> mem v a || mem v b
  | (Tuple a | Ref a) -> mem v a
  | _ -> false

let rec replace x v = function
  | Var i when x = i -> v
  | Fun (a, b) -> Fun (replace x v a, replace x v b)
  | Tuple a -> Tuple (replace x v a)
  | Ref   a -> Ref   (replace x v a)
  | x -> x

let rec expand accu = function
  | [] -> accu
  | (a, b) :: q when a = b -> expand accu q
  | (Fun (a, b), Fun (c, d)) :: q -> expand accu ((a, c) :: (b, d) :: q)
  | ((Tuple a, Tuple b) | (Ref a, Ref b)) :: q -> expand accu ((a, b) :: q)
  | ((Var   x,       t) | (    t, Var x)) :: q ->
      if mem x t then
	raise (Occur_check (x, to_string t))
      else
	expand ((Var x, t) :: accu) q
  | (a, b) :: q -> raise (Clash (to_string a, to_string b))

let rec unify resolved problem =
  match expand [] problem with
    | (Var x, t) :: q ->
	unify
	  (Substitution.add x t (Substitution.map (replace x t) resolved))
          (List.map (fun (a, b) -> (replace x t a, replace x t b)) q)
    | [] -> resolved
    | (a, b) :: q -> raise (Clash (to_string a, to_string b))

let mgu problem = unify Substitution.empty problem

(******************************************************************************************)

let rec generalizable =
  let already = ref [] in
  function expr ->
    try
      List.assq expr !already
    with Not_found -> match expr with
      | VAR _ | FUN _ | INT _ | UNIT _ ->
	  already := (expr, true) :: !already; true
      | (PLUS (a, b) | MINUS (a, b) | TIMES (a, b) | DIV (a, b) | REM (a, b)
	| LESS (a, b) | EQUAL (a, b) | SEQ (a, b) | LET (_, a, b)) ->
	  let res = generalizable a && generalizable b in
	  already := (expr, res) :: !already; res
      | (OPP a | PROJ (a, _) | APP ((VAR "%fix"), a)) ->
	  let res = generalizable a in
	  already := (expr, res) :: !already; res
      | IF (a, b, c) ->
	  let res = generalizable a && generalizable b && generalizable c in
	  already := (expr, res) :: !already; res
      | TUPLE t ->
	  let res = List.fold_left (fun x a -> x && generalizable a) true t in
	  already := (expr, res) :: !already; res
      | _ ->
	  already := (expr, false) :: !already; false

(******************************************************************************************)

let id = Substitution.empty

let ( @ ) f g =
  let g = Substitution.fold (fun x t u -> Substitution.map (replace x t) u) f g in
  Substitution.fold Substitution.add f g

let apply_env f env =
  let aux x t env = TypingEnv.map (fun (q, v) -> (q, replace x t v)) env in
  Substitution.fold aux f env

let apply_type f v = Substitution.fold replace f v
let apply_var f v = try Substitution.find v f with Not_found -> Var v

let rec var_lib = function
  | Var x -> VarSet.add x VarSet.empty
  | Fun (a, b) -> VarSet.union (var_lib a) (var_lib b)
  | (Tuple a | Ref a) -> var_lib a
  | _ -> VarSet.empty

(******************************************************************************************)

let rec subtype env = function

  | VAR x ->
      begin try
	let (bound, t) = TypingEnv.find x env in
	let aux x t = replace x (Var (compteur ())) t in
	let t = VarSet.fold aux bound t in
	(t, id)
      with Not_found ->
	raise (Unbound x)
      end

  | APP (m, n) ->
      let (a, fa) = subtype env m in
      let (b, fb) = subtype (apply_env fa env) n in
      let y = compteur () in
      let mu = mgu [apply_type fb a, Fun (b, Var y)] in
      (apply_var mu y, mu @ fb @ fa)

  | FUN (x, m) ->
      let y = compteur () in
      let env = TypingEnv.add x (VarSet.empty, Var y) env in
      let (a, fa) = subtype env m in
      (Fun (apply_var fa y, a), fa)
      
  | LET (x, m, n) ->
      let (a, fa) = subtype env m in
      let env = apply_env fa env in
      let free =
	if generalizable m then
	  let aux _ (bound, t) free =
	    VarSet.diff free (VarSet.diff (var_lib t) bound)
	  in TypingEnv.fold aux env (var_lib a)
	else
	  VarSet.empty
      in
      let () =
	printf "%s : [ " x;
	VarSet.iter (fun x -> printf "%d " x) free;
	printf "] %s\n" (to_string a);
	flush stdout
      in
      let (b, fb) = subtype (TypingEnv.add x (free, a) env) n in
      (b, fb @ fa)

  | (PLUS (m, n) | MINUS (m, n) | TIMES (m, n) | DIV (m, n) | REM (m, n)) ->
      subtype env (APP (APP (VAR "+", m), n))

  | INT _    -> (Base "int",  id)
  | UNIT ()  -> (Base "unit", id)
  | OPP    m -> subtype env (APP (VAR "-",   m))
  | REF    m -> subtype env (APP (VAR "ref", m))
  | BANG   m -> subtype env (APP (VAR "!",   m))
  | LESS  (m, n) -> subtype env (APP (APP (VAR "<",  m), n))
  | EQUAL (m, n) -> subtype env (APP (APP (VAR "=",  m), n))
  | SET   (m, n) -> subtype env (APP (APP (VAR ":=", m), n))
  | PROJ  (m, n) -> subtype env (APP (APP (VAR ".",  m), n))
  | IF    (m, n, p) -> subtype env (APP (APP (APP (VAR "if", m), n), p))

  | SEQ   (m, n) ->
      let x = compteur () and y = compteur () in
      let env = TypingEnv.add ";" (VarSet.empty, Fun (Var x, Fun (Var y, Var y))) env in
      let (_, fa) as res = subtype env (APP (APP (VAR ";",  m), n)) in
      let b = apply_var fa x in
      let () = if b <> Base "unit" then	fprintf stderr
	"Warning: the expression %s should have type unit but it has type %s.\n"
	(Projet.string_of_expr m) (to_string b); flush stderr
      in res

  | TUPLE vect ->
      let x = compteur () in
      let new_type = List.fold_left (fun a _ -> Fun (Var x, a)) (Tuple (Var x)) vect in
      let new_term = List.fold_left (fun a v -> APP (a, v)) (VAR "[]") vect in
      subtype (TypingEnv.add "[]" (VarSet.empty, new_type) env) new_term

  | CALLPRIM (f, args) ->
      let prim = ";" ^ f in
      let new_term = List.fold_left (fun a v -> APP (a, v)) (VAR prim) args in
      begin
	try
	  ignore (TypingEnv.find prim env);
	  subtype env new_term
	with Not_found ->
	  let x = compteur () in
	  let env = TypingEnv.add prim (VarSet.empty, Var x) env in
	  let (_, fa) as res = subtype env new_term in
	  let () = fprintf stderr
	    "Warning: unknown C primitive \"%s\". It should have type %s.\n"
	    f (to_string (apply_var fa x)); flush stderr
	  in res
      end

  | Op _ -> raise Not_supported

(******************************************************************************************)

let make_base env (x, v, t) =
  TypingEnv.add x
    (List.fold_left (fun bound y -> VarSet.add y bound) VarSet.empty v, t) env

let base_env = List.fold_left make_base TypingEnv.empty
		 [ "+",    [],     Fun (Base "int", Fun (Base "int", Base "int"));
		   "-",    [],     Fun (Base "int", Base "int");
		   "ref",  [0],    Fun (Var 0, Ref (Var 0));
		   "!",    [0],    Fun (Ref (Var 0), Var 0);
		   "<",    [],     Fun (Base "int", Fun (Base "int", Base "bool"));
		   "=",    [0],    Fun (Var 0, Fun (Var 0, Base "bool"));
		   ":=",   [0],    Fun (Ref (Var 0), Fun (Var 0, Base "unit"));
		   ".",    [0],    Fun (Tuple (Var 0), Fun (Base "int", Var 0));
		   "if",   [0],    Fun (Base "bool", Fun (Var 0, Fun (Var 0, Var 0)));
		   "%fix", [0],    Fun (Fun (Var 0, Var 0), Var 0);
		   "afficher_int",  [], Fun   (Base "int", Base "unit");
		   "int_of_string", [], Fun   (Base "string", Base "int");
		   "argv",          [], Tuple (Base "string");
		   "argc",          [], Base "int";
		   "r",             [], Base "string";
		   ";fgetc",   [], Fun  (Base "file", Base "int");
		   ";putchar", [], Fun  (Base "int", Base "unit");
		   ";fopen",   [], Fun  (Base "string", Fun (Base "string", Base "file"));
		   ";fclose",  [], Fun  (Base "file", Base "unit");
		 ]

let type_prog prog =
  let rec aux (LETREC (f, body)) term = match body with
    | FUN (x, t) -> LET (f, APP (VAR "%fix", FUN (f, body)), term)
    | body -> LET (f, body, term)
  in
  let new_term = List.fold_right aux prog (UNIT ()) in
  let _ = subtype base_env new_term in ()

(******************************************************************************************)
